/*
 * Copyright (C) 2001 - 2003 Min Hsu
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <errno.h>
#include <signal.h>

#include <snort.h>
#include <decode.h>
#include <event.h>
#include <plugbase.h>
#include <util.h>

#include <spo_monitor.h>

static inline void sslInit(void)
{
  SSL_library_init();
  SSL_load_error_strings();
}

static inline void sslClean(void)
{
  ERR_remove_state(0);
  ERR_free_strings();
  EVP_cleanup();
}

static void *readCert(const char *pFileName)
{
  FILE *pfd = NULL;
  X509 *pCert = NULL;

  if(!pFileName)
    return NULL;

  if(!(pfd = fopen(pFileName, "r")))
  {
    errorMessage("%s\n", strerror(errno));
    return NULL;
  }

  if(!(pCert = PEM_read_X509(pfd, NULL, NULL, NULL)))
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));

  fclose(pfd);
  return (void *)pCert;
}

static int verifyCert(int iOk, X509_STORE_CTX *pCtx)
{
  X509 *pErrCert = NULL;
  short sErr = 0, sDepth = 0;
  char *pMsg = NULL;

  if(!pCtx) /* Should not be NULL! */
    return 0;

  pErrCert = X509_STORE_CTX_get_current_cert(pCtx);
  sErr = X509_STORE_CTX_get_error(pCtx);
  sDepth=  X509_STORE_CTX_get_error_depth(pCtx);

#ifdef DEBUG
  pMsg = X509_NAME_oneline (X509_get_issuer_name(pErrCert), 0, 0);
  debugMessage("Issuer: %s\n", pMsg);
  free(pMsg);

  pMsg = X509_NAME_oneline(X509_get_subject_name(pErrCert), 0, 0);
  debugMessage("Subject: %s\n", pMsg);
  free(pMsg);
#endif

  if(iOk)
  {
    debugMessage("Verify error: %d:%s\n", sErr,
                            X509_verify_cert_error_string(sErr));
    if(sDepth <= CHAINDEPTH)
      iOk=1;
    else
    {
      iOk=0;
      X509_STORE_CTX_set_error(pCtx, X509_V_ERR_CERT_CHAIN_TOO_LONG);
    }
  }

  switch (pCtx->error)
  {
    case X509_V_ERR_CERT_CHAIN_TOO_LONG:
         errorMessage("Certificate chain is too long!\n");
         break;

    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
         pMsg = X509_NAME_oneline(
                X509_get_issuer_name(pCtx->current_cert), 0 ,0);

         errorMessage("Unable to get CA: %s\n", pMsg);
         free(pMsg);
         iOk=0;
         break;

    case X509_V_ERR_CERT_NOT_YET_VALID:
    case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
         errorMessage("Certificate not valid yet!\n");
         iOk=0;
         break;

    case X509_V_ERR_CERT_HAS_EXPIRED:
    case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
         errorMessage("Certificate has expired!\n");
         iOk=0;
         break;
  }

  return(iOk);
}

static SSL_CTX *initClientCtx(void *pCert)
{
  SSL_CTX *pCtx = NULL;
  SSL_METHOD *pMthd = NULL;

  if(!pCert)
    return NULL;

  pMthd = SSLv3_client_method();

  pCtx = SSL_CTX_new(pMthd);
  if(!pCtx)
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
    return NULL;
  }

  SSL_CTX_set_options(pCtx, SSL_OP_NO_SSLv2 | SSL_OP_NO_TLSv1);

  SSL_CTX_set_verify_depth(pCtx, CHAINDEPTH);
  SSL_CTX_set_verify(pCtx, SSL_VERIFY_PEER , verifyCert);

  if(!X509_STORE_add_cert(pCtx->cert_store, pCert))
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));

  return pCtx;
}

static inline void cleanCtx(SSL_CTX *pCtx)
{
  if(pCtx)
    SSL_CTX_free(pCtx);
}

static SSL *connetServer(SSL_CTX *pCtx, const int iSock)
{
  SSL *pConnect = NULL;

  if(!pCtx || iSock <= 0)
    return NULL;

  if(!(pConnect = SSL_new(pCtx)))
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
    return NULL;
  }

  if(SSL_set_fd(pConnect, iSock) > 0)
  {
    if(SSL_connect(pConnect) > 0) /* Connect to server */
      return pConnect;
    else
    {
      if(ERR_get_error())
        errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      else
        errorMessage("%s\n", strerror(errno));
    }
  }
  else
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));

  SSL_free (pConnect);
  return NULL;
}

static inline void disconnectSSL(SSL *pConnect)
{
  if(!pConnect)
    return;

  SSL_shutdown(pConnect);
  SSL_free(pConnect);
}

static inline int uTimeWait(const int iUSec)
{
  int iErr = 0;
  struct timeval tOut = {0, 0};

  tOut.tv_usec = iUSec;

  while((iErr = select(0, NULL, NULL, NULL, &tOut)))
    if(errno != EINTR)
      break;

  return iErr;
}

static inline void buildEvtMsg(unsigned char *pBuff, SNSREVENT *pEvt)
{
  if(!pBuff || !pEvt)
    return;

  memmove(pBuff, pEvt, PSEUDOHDLEN);
  memmove(pBuff+PSEUDOHDLEN, pEvt->pEventMsg, ntohs(pEvt->usEvtLen));
}

static void buildSttMsg(unsigned char *pBuff, SNSRMONIT *pMonit)
{
  if(!pBuff || !pMonit)
    return;
  
  /* Build statistic message from _PacketCount structure */
#define STATIS pMonit->pStats
  memset(STATIS, 0, sizeof(SNSRSTATS));

  STATIS->ulMsgId = htonl(STATIMSG);
  STATIS->ulSensorId = htonl(pMonit->pCfg->ulSid);
  STATIS->ulAltPkts = pc.alert_pkts;
  STATIS->ulLogPkts = pc.log_pkts;
  STATIS->ulPsdPkts = pc.pass_pkts;
  STATIS->ulTcpPkts = pc.tcp;
  STATIS->ulUdpPkts = pc.udp;
  STATIS->ulIcmpPkts = pc.icmp;
  STATIS->ulArpPkts = pc.arp;
  STATIS->ulIpxPkts = pc.ipx;
  STATIS->ulOthPkts = pc.other;
  STATIS->ulDscdPkts = pc.discards;
  STATIS->ulFrgIPPkts = pc.frags;
  STATIS->ulRbtIpPkts = pc.rebuilt_frags;
  STATIS->ulStrmTcpPkts = pc.tcp_stream_pkts;
  STATIS->ulRcnsTcpPkts = pc.rebuilt_tcp;
  STATIS->ulRcnsStrmPkts = pc.tcp_streams;

  if(gettimeofday(&STATIS->sensorTime, NULL))
    errorMessage("%s\n", strerror(errno));

  memcpy(pBuff, STATIS, sizeof(SNSRSTATS));
#undef STATIS
}

static int readSSL(SSL *pConnect, unsigned char *pData, const int iLen)
{
  int iDataLen = iLen;
  int iDataRcv = 0;
  int iRcvLen = 0;

  if(!pConnect || !pData)
    return -1;

  while(iDataLen > 0)
  {
    iDataRcv = SSL_read(pConnect, pData+iRcvLen, iLen);

    switch(SSL_get_error(pConnect, iDataRcv))
    {
      case SSL_ERROR_NONE:
             if(SSL_pending(pConnect))
             {
               iRcvLen += iDataRcv;
               iDataLen -= iDataRcv;
               break;
             }
             else
               return iDataRcv;

      case SSL_ERROR_WANT_READ:
      case SSL_ERROR_WANT_WRITE:
      case SSL_ERROR_WANT_X509_LOOKUP:
             uTimeWait(1000);
             break;

      default:
             errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
             return -1;
    }
  }

  return iRcvLen;
}

static int writeSSL(SSL *pConnect, const unsigned char *pData, const int iLen)
{
  int iDataLen = iLen;
  int iDataSnd = 0;
  int iSndLen = 0;

  if(!pConnect || !pData)
    return -1;

  while(iDataLen > 0)
  {
    iDataSnd = SSL_write(pConnect, pData+iSndLen, iLen);
    switch(SSL_get_error(pConnect, iDataSnd))
    {
      case SSL_ERROR_NONE:
             iSndLen += iDataSnd;
             iDataLen -= iDataSnd;
             break;
      case SSL_ERROR_WANT_WRITE:
      case SSL_ERROR_WANT_READ:
      case SSL_ERROR_WANT_X509_LOOKUP:
             uTimeWait(1000);
             break;
      default:
             errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
             return -1;
    }
  }

  return iSndLen;
}

static inline int mtxLock(pthread_mutex_t *pMtx)
{
  if(pMtx)
    return pthread_mutex_lock(pMtx);
  else
    return -1;
}

static inline int mtxUnLock(pthread_mutex_t *pMtx)
{
  if(pMtx)
    return pthread_mutex_unlock(pMtx);
  else
    return -1;
}

static inline int cndSignal(pthread_cond_t *pCnd)
{
  if(pCnd)
    return pthread_cond_signal(pCnd);
  else
    return -1;
}

static inline int cndTimedWait(pthread_cond_t *pCnd,
                               pthread_mutex_t *pMtx, unsigned long millis)
{
  int iErr=0;
  struct timeval tv;
  struct timespec ts;

  if(!pCnd || !pMtx)
    return -1;

  if(gettimeofday(&tv, NULL))
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }

  ts.tv_sec = tv.tv_sec + (millis/1000);
  ts.tv_nsec = tv.tv_usec * 1000;

  while((iErr=pthread_cond_timedwait(pCnd, pMtx, &ts))==EINTR);

  return iErr;
}

static inline EVENTPOOL *newEvent(EVENTPOOL *pEvt)
{
   EVENTPOOL *pLast = NULL;

  if(!pEvt)
    return NULL;

  GETLAST(pEvt, pLast);

  /* Allocte a new one */
  if(!(pLast->pNext = (EVENTPOOL *)calloc(1, sizeof(EVENTPOOL))))
  {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
    errorMessage("%s\n", strerror(errno));
#else
    errorMessage("calloc failed.\n");
#endif

    return NULL;
  }

  pLast->pNext->pPrev = pLast;

  return pLast->pNext;
}

static inline void deleteEvent(EVENTPOOL *pEvt)
{
  if(!pEvt)
    return;

  if(pEvt->pPrev)
    pEvt->pPrev->pNext = pEvt->pNext;

  if(pEvt->pNext)
    pEvt->pNext->pPrev = pEvt->pPrev;

  if(pEvt->pEvent)
  {
    if(pEvt->pEvent->pEventMsg)
      free(pEvt->pEvent->pEventMsg);

    free(pEvt->pEvent);
  }

  free(pEvt);
}

static inline void deleteAllEvt(EVENTPOOL *pEvt)
{
  EVENTPOOL *pFirst = NULL;

  if(!pEvt)
    return;

  GETFIRST(pEvt, pFirst);

  while(pFirst->pNext)
  {
    pFirst = pFirst->pNext;
    deleteEvent(pFirst->pPrev);
  }

  deleteEvent(pFirst);
}

static int connectNonBlock(int iSock, const char *pIP,
                           const int iPort, const short sTimeout)
{
  int iRc = 0;
  int iFlag = 0;
  int iErr = 0;
  socklen_t tLen = sizeof(iErr);
  struct sockaddr_in skin;
  fd_set  tRset, tWset;
  struct timeval tWait;

  if((iFlag = fcntl(iSock, F_GETFL, 0)) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    return iFlag;
  }

  if(fcntl(iSock, F_SETFL, iFlag|O_NONBLOCK) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }

  memset(&skin, 0, sizeof(skin));

  skin.sin_family = AF_INET;
  skin.sin_port = htons(iPort);

  if(!inet_aton(pIP, &skin.sin_addr))
  {
    errorMessage("inet_aton failed ...\n");
    return -1;
  }
  
  iRc = connect(iSock, (struct sockaddr *)&skin, sizeof(skin));

  if(iRc == -1 && errno != EINPROGRESS)
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }
  else
  {
    tWait.tv_sec = sTimeout;
    tWait.tv_usec = 0;
    FD_ZERO(&tRset);
    FD_SET(iSock, &tRset);
    tWset = tRset;

    switch(select(iSock+1, &tRset, &tWset, (fd_set *)0, &tWait))
    {
      case -1:
              errorMessage("%s\n", strerror(errno));
              return -1;
      case  0:
              errno = ETIMEDOUT; /* No network traffic, time out */
              return -1;
      default:
              if(FD_ISSET(iSock, &tRset) || FD_ISSET(iSock, &tWset))
              {
                if(getsockopt(iSock, SOL_SOCKET, SO_ERROR, &iErr, &tLen) == -1)
                {
                  errorMessage("%s\n", strerror(errno));
                  return -1;
                }

                if(iErr)
                {
                  errno = iErr;
                  errorMessage("%s\n", strerror(errno));
                  return -1;
                }
              }
    }
  } /* else */

  if(fcntl(iSock, F_SETFL, iFlag) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }

  return 0;
}

static void *sensorMonitorThread(void *pUser)
{
  SNSRMONIT *pMonit = (SNSRMONIT *)pUser;
  X509 *pCert = NULL;
  SSL_CTX *pCtx = NULL;
  SSL *pCon = NULL;
  time_t tRfncTime = 0;

#define CNFG pMonit->pCfg
#define ENVET pMonit->pEvent
    
  sslInit();

  while(pMonit && !pMonit->cException)
  {
    debugMessage("Sensor monitor thread starting ...\n");

    if(!(pCert = readCert(CNFG->pCertificate)))
    {
      errorMessage("readCert failed to read %s ...\n", CNFG->pCertificate);
      break;
    }

    if(!(pCtx = initClientCtx(pCert)))
    {
      errorMessage("%s\n", "initClientCtx failed ...\n");
      break;
    }

    while(!pMonit->cException)
    {
      int iOpt = 1;

      if((pMonit->iSockFd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
      {
        errorMessage("%s\n", strerror(errno));
        pMonit->cException = 1;
        break;
      }

      if(setsockopt(pMonit->iSockFd, SOL_SOCKET, SO_REUSEADDR,
                        (char *)&iOpt, sizeof(iOpt)) == -1)
      {
        errorMessage("%s\n", strerror(errno));
        pMonit->cException = 1;
        break;
      }

      /* Set 5 seconds timeout. */
      if(!connectNonBlock(pMonit->iSockFd, CNFG->mntrIP, CNFG->usPort, 5))
      {
        debugMessage("Connected to monitor!\n");

        /* Start SSL handshake. */
        if(!(pCon = connetServer(pCtx, pMonit->iSockFd)))
        {
          debugMessage("SSL handshake failed!\n");
          close(pMonit->iSockFd);
          pMonit->cException = 1;
          break;
        }
        /* SSL verification successfully finished. */
        debugMessage("SSL handshake successed!\n");
      }
      else
      {
        debugMessage("Connect to monitor failed.\n");
        close(pMonit->iSockFd);
        /* Wait some time before to re-try connection. */
        uTimeWait(20000000);
        continue;
      }

      while(!pMonit->cException)
      {
        EVENTPOOL *pFirst = NULL;
        time_t tCrntTime = time(0);

        if(tCrntTime - tRfncTime >= MSGTIMEINTV)
        {
          unsigned char *pBuff = NULL;
          short sSndLen = 0;
          unsigned short usBuffLen = sizeof(SNSRSTATS);
          
          pBuff = (unsigned char *)calloc(1, usBuffLen);

          if(!pBuff)
          {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
            errorMessage("%s!\n", strerror(errno));
#else
            errorMessage("calloc failed ...\n");
#endif
            pMonit->cException = 1;
            break;
          }
   
          buildSttMsg(pBuff, pMonit);

          debugMessage("Statistic message built ...\n");

          sSndLen = writeSSL(pCon, pBuff, usBuffLen);

          free(pBuff);

          if(sSndLen != usBuffLen)
          {
            errorMessage("writeSSL failed ...\n");
            break;
          }

          sSndLen = readSSL(pCon, &pMonit->ucCtrlMsg, 1);

          if(sSndLen == -1)
          {
            errorMessage("readSSL failed ...\n");
            break;
          }

          debugMessage("Get control message: %d\n", pMonit->ucCtrlMsg);

          tRfncTime = tCrntTime; 
        }

        if(pMonit->ucCtrlMsg == SNSRSTOP) /* Stop the sensor */
        {
          pMonit->cException = 1;
          break;
        }

        mtxLock(&pMonit->tMutex);

        GETFIRST(pMonit->pEvtPl, pFirst); 

#define EVENT pFirst->pNext->pEvent

        if(!pFirst->pNext || !EVENT)
          /* Block and wait data ready. */
          cndTimedWait(&pMonit->tCond, &pMonit->tMutex, 10000);
        else
        {
          unsigned char *pBuff = NULL;
          short sSndLen = 0;
          unsigned short usBuffLen = PSEUDOHDLEN + ntohs(EVENT->usEvtLen);

          debugMessage("Sending event ...\n");

          pBuff = (unsigned char *)calloc(1, usBuffLen);

          if(!pBuff)
          {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
            errorMessage("%s!\n", strerror(errno));
#else
            errorMessage("calloc failed ...\n");
#endif
            pMonit->cException = 1;
            mtxUnLock(&pMonit->tMutex);
            break;
          }

          buildEvtMsg(pBuff, EVENT);

#undef EVENT

          debugMessage("Event message built ...\n");

          sSndLen = writeSSL(pCon, pBuff, usBuffLen);
          deleteEvent(pFirst->pNext);

          free(pBuff);

          if(sSndLen != usBuffLen)
          {
            errorMessage("writeSSL failed ...\n");
            mtxUnLock(&pMonit->tMutex);
            break;
          }
          debugMessage("Event message sent %d ...\n", sSndLen);
        } /* else */

        mtxUnLock(&pMonit->tMutex);
      } /* while */
    } /* while */  

    disconnectSSL(pCon);
    cleanCtx(pCtx);
    X509_free(pCert);
    close(pMonit->iSockFd);
    break;
  } /* while */

#undef CNFG
#undef ENVET

  sslClean();

  pMonit->tSpndId = 0;

  if(pMonit->ucCtrlMsg == SNSRSTOP)
    pthread_kill(pMonit->tMainId, SIGTERM);

  debugMessage("Sensor monitor thread stopping ...\n");

  return NULL;
}

static void snortMonitor(Packet *pPkt, char *pMsg, void *pUser, Event *pEvt)
{
  SNSRMONIT *pMonit = (SNSRMONIT *)pUser;
  SNSREVENT *pNewEvt = NULL;

  debugMessage("Monitor Event!\n");

  if(!pPkt && !pMsg)
    return;

  if(pMonit && !pMonit->cException && pMonit->ucCtrlMsg == SENSORUP)
  {
    mtxLock(&pMonit->tMutex);

    if(!(pNewEvt = (SNSREVENT *)calloc(1, sizeof(SNSREVENT))))
    {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
      errorMessage("%s!\n", strerror(errno));
#else
      errorMessage("calloc failed ...\n");
#endif
      return;
    }

    pNewEvt->ulMsgId = htonl(EVENTMSG);
    pNewEvt->ulSgntrId = htonl(pEvt->sig_id);
    pNewEvt->ulSensorId = htonl(pMonit->pCfg->ulSid);

    if(pPkt)
    {
      if(pPkt->eh)
      {
        if(ntohs(pPkt->eh->ether_type) == 0x0800) /* ETHERTYPE_IP */
          pNewEvt->usPrtclID = pPkt->iph->ip_proto; 
        else
          pNewEvt->usPrtclID = pPkt->eh->ether_type;

        *pNewEvt->ucSrcMac = *pPkt->eh->ether_src; 
        *(pNewEvt->ucSrcMac+1) = *(pPkt->eh->ether_src+1); 
        *(pNewEvt->ucSrcMac+2) = *(pPkt->eh->ether_src+2); 
        *(pNewEvt->ucSrcMac+3) = *(pPkt->eh->ether_src+3); 
        *(pNewEvt->ucSrcMac+4) = *(pPkt->eh->ether_src+4); 
        *(pNewEvt->ucSrcMac+5) = *(pPkt->eh->ether_src+5); 
        *pNewEvt->ucDstMac = *pPkt->eh->ether_dst;
        *(pNewEvt->ucDstMac+1) = *(pPkt->eh->ether_dst+1);
        *(pNewEvt->ucDstMac+2) = *(pPkt->eh->ether_dst+2);
        *(pNewEvt->ucDstMac+3) = *(pPkt->eh->ether_dst+3);
        *(pNewEvt->ucDstMac+4) = *(pPkt->eh->ether_dst+4);
        *(pNewEvt->ucDstMac+5) = *(pPkt->eh->ether_dst+5);
      }
      else
        pNewEvt->usPrtclID = pPkt->iph->ip_proto; 

      pNewEvt->ulSrcIp = pPkt->iph->ip_src.s_addr;
      pNewEvt->ulDstIp = pPkt->iph->ip_dst.s_addr;

      switch(pNewEvt->usPrtclID)
      {
        case IPPROTO_TCP:
             pNewEvt->usSrcPort = pPkt->tcph->th_sport;
             pNewEvt->usDstPort = pPkt->tcph->th_dport;
             break;

        case IPPROTO_UDP:
             pNewEvt->usSrcPort = pPkt->udph->uh_sport;
             pNewEvt->usDstPort = pPkt->udph->uh_dport;
             break;

        case IPPROTO_ICMP:
             if(pPkt->icmph)
             {
               pNewEvt->usSrcPort = htons((unsigned short)pPkt->icmph->type);
               pNewEvt->usDstPort = htons((unsigned short)pPkt->icmph->code);
             }
      }

      (pNewEvt->eventTime).tv_sec = (pPkt->pkth->ts).tv_sec;
      (pNewEvt->eventTime).tv_usec = (pPkt->pkth->ts).tv_usec;
    }
    else
      gettimeofday(&pNewEvt->eventTime, NULL);

    if(pEvt->priority)
      pNewEvt->usPriority = htons(pEvt->priority);

    pNewEvt->usEvtLen = htons(strlen(pMsg));
    pNewEvt->pEventMsg = strdup(pMsg);

    (newEvent(pMonit->pEvtPl))->pEvent = pNewEvt;

    /* Signal the sending thread */
    cndSignal(&pMonit->tCond);

    mtxUnLock(&pMonit->tMutex);
  }
}

static char *getConfigData(const char *pArgs, const char *pTag)
{
  short totalLen = 0;
  short itemLen = 0;
  short tagLen = 0;
  char *pFound = NULL;
  char *pSpace = NULL;
  char *pData = NULL;

  if(!pArgs || !pTag)
    return NULL;

  totalLen = strlen(pArgs);
  tagLen = strlen(pTag);

  pFound = strstr(pArgs, pTag);

  if(!pFound)
    return NULL;

  pSpace = index(pFound, 0x20);

  if(pSpace)
    itemLen = (short)(pSpace - pFound) - 1;
  else
    itemLen = totalLen - (short)(pFound - pArgs);

  if(!(pData = (char *)calloc(1, itemLen-tagLen)))
    return NULL;

  memmove(pData, pFound+tagLen+1, itemLen-tagLen);

  return pData;
}

static CONFGINFO *parseConfigArgs(char *pArgs)
{
  CONFGINFO *pCfg = NULL;
  char *pTmp = NULL;

  if(!pArgs)
    errorMessage("Oops! No config info supplied, use default.\n");

  pCfg = (CONFGINFO *)calloc(1, sizeof(CONFGINFO));

  if(!pCfg)
  {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
    errorMessage("%s\n", strerror(errno));
#else
    errorMessage("calloc failed ...\n");
#endif
    return NULL;
  }

  pTmp = getConfigData(pArgs, "ipaddr");

  if(!pTmp)
    /* No monitor IP address found, assume localhost. */
    memcpy(pCfg->mntrIP, "127.0.0.1", 9);
  else
  {
    memcpy(pCfg->mntrIP, pTmp, strlen(pTmp));
    free(pTmp);
  }

  pTmp = getConfigData(pArgs, "port");

  if(!pTmp)
    /* No port number found, use default port. */
    pCfg->usPort = 2001;
  else
  {
    int i = atoi(pTmp);
    pCfg->usPort = (i<=0 || i >0xFFFF)? 2001 : i;
    free(pTmp);
  }

  pTmp = getConfigData(pArgs, "snsrid");

  if(!pTmp)
    /* No sensor id found, set id to 0x001. */
    pCfg->ulSid = 0x001;
  else
  {
    int l = atoi(pTmp);
    pCfg->ulSid = (l<=0)? 0x001 : l;
    free(pTmp);
  }

  pTmp = getConfigData(pArgs, "cert");

  if(!pTmp)
    /* No certificate file found, assume current directory. */
    pCfg->pCertificate = strdup("./cert.pem");
  else
  {
    pCfg->pCertificate = strdup(pTmp);
    free(pTmp);
  }

  return pCfg;
}

static void cleanUpMonitor(int iRcvSig, void *pUser)
{
  SNSRMONIT *pSensorMonitor = pUser;
  debugMessage("Sensor monitor cleaning up ...\n");

  if(pSensorMonitor)
  {
    if(pSensorMonitor->tSpndId)
    {
      pSensorMonitor->cException = 1;
      /* Signal the sending thread */
      cndSignal(&pSensorMonitor->tCond);
      pthread_join(pSensorMonitor->tSpndId, NULL);
    }

    if(pSensorMonitor->iSockFd != -1)
    {
      debugMessage("Close socket ...\n");
      close(pSensorMonitor->iSockFd);
    }

    if(pSensorMonitor->pCfg)
    {
      if(pSensorMonitor->pCfg->pCertificate)
        free(pSensorMonitor->pCfg->pCertificate);
      free(pSensorMonitor->pCfg);
    }

    deleteAllEvt(pSensorMonitor->pEvtPl);

    if(pSensorMonitor->pStats)
      free(pSensorMonitor->pStats);

    pthread_mutex_destroy(&pSensorMonitor->tMutex);
    pthread_cond_destroy(&pSensorMonitor->tCond);

    free(pSensorMonitor);
  }
  debugMessage("Sensor monitor cleaned ...\n");
}

static void initSnortMonitor(unsigned char *pArgs)
{
  int iErr = 0;
  SNSRMONIT *pSensorMonitor = NULL;
  
  debugMessage("Sensor monitor activated!\n");

  pSensorMonitor = (SNSRMONIT *)calloc(1, sizeof(SNSRMONIT));

  if(!pSensorMonitor)
  {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
    errorMessage("%s\n", strerror(errno));
#else
    errorMessage("calloc failed ...\n");
#endif
  }

  while(pSensorMonitor)
  {
    pSensorMonitor->pStats = (SNSRSTATS *)malloc(sizeof(SNSRSTATS));

    if(!pSensorMonitor->pStats)
    {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
      errorMessage("%s\n", strerror(errno));
#else
      errorMessage("calloc failed ...\n");
#endif
      break;
    }

    pSensorMonitor->pEvtPl = (EVENTPOOL *)calloc(1, sizeof(EVENTPOOL));

    if(!pSensorMonitor->pEvtPl)
    {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
      errorMessage("%s\n", strerror(errno));
#else
      errorMessage("calloc failed ...\n");
#endif
      break;
    }

    pSensorMonitor->pCfg = parseConfigArgs(pArgs);

    if(!pSensorMonitor->pCfg)
    {
      errorMessage("parseConfigArgs failed ...\n");
      break;
    }

    /* pthread_mutex_init always returns 0. */
    pthread_mutex_init(&pSensorMonitor->tMutex, NULL);

    if(pthread_cond_init(&pSensorMonitor->tCond, NULL))
    {
      errorMessage("pthread_cond_init failed ...\n");
      break;
    }

    pSensorMonitor->iSockFd = -1;
    pSensorMonitor->tMainId = pthread_self();

    if((iErr=pthread_create(&pSensorMonitor->tSpndId, NULL,
                          sensorMonitorThread, (void *)pSensorMonitor)))
    {
      errorMessage("%s\n", strerror(iErr));
      break;
    }
  
    pv.alert_plugin_active = 1;

    AddFuncToOutputList(snortMonitor, NT_OUTPUT_ALERT, pSensorMonitor);
    AddFuncToCleanExitList(cleanUpMonitor, pSensorMonitor); 

    return;
  }

  if(pSensorMonitor)
  {
    if(pSensorMonitor->pStats)
      free(pSensorMonitor->pStats);
    if(pSensorMonitor->pEvtPl)
      free(pSensorMonitor->pEvtPl);

    free(pSensorMonitor);
  }

  FatalError("Snort monitor initialization failed ...\n");
}

void SetupMonitor()
{
  RegisterOutputPlugin("alert_monitor", NT_OUTPUT_ALERT, initSnortMonitor);
  debugMessage("Register SetupMonitor done.\n");
}
