/*
 * Copyright (C) 2001 - 2003 Min Hsu
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <openssl/asn1.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/pem.h>

#include <gencerts.h>
#include <loadkeys.h>
#include <printmsg.h>

static unsigned long long getRandNum(void)
{
  int iFd = 0;
  unsigned long long ullRandNum = 0;

  if((iFd = open("/dev/urandom", O_RDONLY)) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    return 0;
  }

  if(read(iFd, (char *)&ullRandNum, 8) == -1) 
  {
    errorMessage("%s\n", strerror(errno));
    ullRandNum = 0;
  }

  close(iFd);

  return ullRandNum;
}

/* Subject: SystemName:User@host:RandomHex */ 

static char *buildName(void) 
{
  struct passwd *pPW = NULL;
  struct utsname uts;
  char *pName = NULL;
  short sNameLen = 0;
#ifdef _REENTRANT
  struct passwd tPW;
  char rslt[4096];
#endif
        
#ifdef _REENTRANT
  /*FIXME get length of the /etc/passwd instead of 4096 */
  getpwuid_r(getuid(), &tPW, rslt, 4096, &pPW);
#else
  pPW = getpwuid(getuid());
#endif


  if(!pPW)
  {
    errorMessage("%s\n", strerror(errno));
    return NULL;
  }

  if(pPW->pw_name)
    sNameLen += strlen(pPW->pw_name);

  if(uname(&uts))
  {
    errorMessage("%s\n", strerror(errno));
    return NULL;
  }

  sNameLen += strlen(uts.nodename);
  sNameLen += strlen(uts.sysname);
  sNameLen += 3;

  pName = (char *)calloc(1, sNameLen+16+1);
  snprintf(pName, sNameLen+16+1, "%s:%s@%s:%llX", uts.sysname, pPW->pw_name, 
                                 uts.nodename, getRandNum());
        
  return pName;
}

static int setReqSubject(X509_REQ *pReq)
{
  X509_NAME *pName = NULL;
  char *pSubName = NULL;
  short rc = 0;

  while(pReq)
  {
    pName = pReq->req_info->subject;

    /* setup version number */
    if(!X509_REQ_set_version(pReq, 0L))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    pSubName = buildName();

    debugMessage("Subject: %s\n", pSubName);

    rc = X509_NAME_add_entry_by_NID(pName, OBJ_txt2nid("CN"), MBSTRING_ASC,
                                   (unsigned char *)pSubName, -1, -1, 0);
    free(pSubName);

    if(!rc)
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!X509_NAME_entry_count(pName))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    return 0;
  }

  return -1;
}

static X509_REQ *createX509Req(EVP_PKEY *pKey, const EVP_MD *pHash)
{
  X509_REQ *pReq = NULL;

  while(pKey && pHash)
  {
    /* Create a new REQ */
    pReq = X509_REQ_new();
    if(!pReq)
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(setReqSubject(pReq))
    {
      errorMessage("%s\n", "setReqSubject fail ...");
      break;
    }

    X509_REQ_set_pubkey(pReq, pKey);

    /* Sign the new REQ */
    if(!X509_REQ_sign(pReq, pKey, pHash))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    return pReq;
  }

  if(pReq)
    X509_REQ_free(pReq);

  return NULL;
}

static X509 *signCertReq(RTCACERT *pCa, EVP_PKEY *pPkey, X509_REQ *pReq,
                         const EVP_MD *pHash, short sDays)
{
  X509 *pX509sc = NULL;

  while(pCa && pReq && pHash && pPkey)
  {
    if(X509_REQ_verify(pReq, pPkey) <= 0)
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!(pX509sc = X509_new()))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    /* Set version to V3 */
    if(!X509_set_version(pX509sc, 2))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!ASN1_INTEGER_set(X509_get_serialNumber(pX509sc), 0))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!X509_set_issuer_name(pX509sc, pReq->req_info->subject))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if (!X509_set_subject_name(pX509sc, pReq->req_info->subject))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    /* FIXME should adjust to match remote time zone */
    X509_gmtime_adj(X509_get_notBefore(pX509sc), -14400);
#define SECINDAYS 60*60*24*sDays
    X509_gmtime_adj(X509_get_notAfter(pX509sc), (long)SECINDAYS);
#undef SECINDAYS

    X509_set_pubkey(pX509sc, pPkey);

    if(!X509_check_private_key(pCa->pRootCA, pCa->pRootPrivKey))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }
  
    if(!ASN1_INTEGER_set(X509_get_serialNumber(pX509sc), 1))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!X509_set_issuer_name(pX509sc, X509_get_subject_name(pCa->pRootCA)))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!X509_sign(pX509sc, pCa->pRootPrivKey, pHash))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    return pX509sc;
  }

  if(pX509sc)
    X509_free(pX509sc);

  return NULL;
}

SGNDCERT *createX509Cert(RTCACERT *pCert, const short sKeySize,
                                          const short sDays)
{
  SGNDCERT *pNewCert = NULL;
  X509_REQ *pReq = NULL;
  const EVP_MD *pHash = EVP_md5();
  short sKeyLen = 0;

  if(!EVP_get_digestbyname("EVP_md5"))
  {
    if(!EVP_add_digest(EVP_md5()))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      return NULL;
    }
  }

  while(pCert)
  {
    sKeyLen = (sKeySize < DEFAULTKLEN)? DEFAULTKLEN : sKeySize;

    pNewCert = (SGNDCERT *)calloc(1, sizeof(SGNDCERT));

    if(!pNewCert)
    {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
      errorMessage("%s\n", strerror(errno));
#else
      errorMessage("calloc failed.\n");
#endif

      break;
    }

    debugMessage("Generating a %d bit %s private key\n", sKeySize, "RSA");

    if(!(pNewCert->pPkey = EVP_PKEY_new()))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    if(!EVP_PKEY_assign_RSA(pNewCert->pPkey, 
                            RSA_generate_key(sKeyLen, 0x10001, NULL, NULL)))
    {
      errorMessage("%s\n", ERR_reason_error_string(ERR_get_error()));
      break;
    }

    pReq = createX509Req(pNewCert->pPkey, pHash);

    if(!pReq)
    {
      errorMessage("%s\n", "createX509Req fail ...");
      break;
    }

    if(!(pNewCert->pSignedCert = signCertReq(pCert, pNewCert->pPkey,
                                                    pReq, pHash, sDays)))
    {
      errorMessage("%s\n", "signCertReq fail ...");
      break;
    }

    X509_REQ_free(pReq);
    return pNewCert;
  }

  if(pReq)
    X509_REQ_free(pReq);

  if(pNewCert)
  {
    if(pNewCert->pPkey)
      EVP_PKEY_free(pNewCert->pPkey);

    free(pNewCert);
  }

  return NULL;
}

RTCACERT *loadCaCert(const char *pCertFile, const char *pPrvkFile)
{
  RTCACERT *pCert = NULL;

  pCert = (RTCACERT *)calloc(1, sizeof(RTCACERT));

  if(!pCert)
  {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
    errorMessage("%s\n", strerror(errno));
#else
    errorMessage("calloc failed.\n");
#endif

    return NULL;
  }

  while(pCert)
  {
    if(!(pCert->pRootCA = readCert(pCertFile)))
    {
      errorMessage("%s\n", "readCert fail ...");
      break;
    }

    if(!(pCert->pRootPrivKey = (EVP_PKEY *)readPrivKey(pPrvkFile)))
    {
      errorMessage("%s\n", "readPrivKey fail ...");
      break;
    }

    return pCert;
  }

  if(pCert)
    free(pCert);

  return NULL;
}

void freeCaCert(RTCACERT *pCert)
{
  if(pCert)
  {
    if(pCert->pRootCA)
      X509_free(pCert->pRootCA);
    if(pCert->pRootPrivKey)
      EVP_PKEY_free(pCert->pRootPrivKey);

    free(pCert);
  }

  ERR_remove_state(0);
  ERR_free_strings();
  EVP_cleanup();
}

void freeSignedCert(SGNDCERT *pCert)
{
  if(pCert)
  {
    if(pCert->pSignedCert)
      X509_free(pCert->pSignedCert);
    if(pCert->pPkey)
      EVP_PKEY_free(pCert->pPkey);

    free(pCert);
  }
}
