/*
 * Copyright (C) 2001 - 2003 Min Hsu
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <errno.h>

#include <printmsg.h>
#include <gencerts.h>
#include <sslutils.h>
#include <gmlmonit.h>
#include <dbprocs.h>

/* Global data */
MONITSRVR tMonitSrv;

static int openSocket(const unsigned short usPort, const int iBackLog)
{
  int iOpt = 1;
  int iSockFd = -1;
  struct sockaddr_in sin;

  if((iSockFd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = htons(usPort);

  if(setsockopt(iSockFd, SOL_SOCKET, SO_REUSEADDR,
                             (void *)&iOpt, sizeof(iOpt)) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    close(iSockFd);
    return -1;
  }

  if(bind(iSockFd, (struct sockaddr *)&sin, sizeof(sin)))
  {
    errorMessage("%s\n", strerror(errno));
    close(iSockFd);
    return -1;
  }

  if(listen(iSockFd, iBackLog) == -1)
  {
    errorMessage("%s\n", strerror(errno));
    close(iSockFd);
    return -1;
  }

  /* Set the socket to nonblock mode */
  if(ioctl(iSockFd, FIONBIO, &iOpt))
  {
    errorMessage("%s\n", strerror(errno));
    return -1;
  }

  return(iSockFd);
}

static void signalQUT(int iSigNum)
{
  debugMessage("signalQUT, Received signal: %d\n", iSigNum);
  tMonitSrv.cException = 1;
}

static void signalHandler(void)
{
  struct sigaction actIGN;
  struct sigaction actQUT;

  sigemptyset(&actIGN.sa_mask);
  actIGN.sa_flags = SA_NODEFER;
  actIGN.sa_handler = SIG_IGN;

  sigemptyset(&actQUT.sa_mask);
  actQUT.sa_flags = SA_ONESHOT;
  actQUT.sa_handler = signalQUT;

  if(sigaction(SIGHUP, &actQUT, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGQUIT, &actQUT, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGTERM, &actQUT, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGINT, &actIGN, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGALRM, &actIGN, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGPIPE, &actIGN, NULL))
    errorMessage("%s\n", strerror(errno));

  if(sigaction(SIGCHLD, &actIGN, NULL))
    errorMessage("%s\n", strerror(errno));
}

int main(int agrc, char **argv)
{
  int iSockFd = -1;
  int iPort = 0;
  char *pRootCa = NULL;
  char *pCaPrvKey = NULL;
  char *pCnfgFile = NULL;
  MONITSRVR *pMntSrv = &tMonitSrv;

  if(getuid() != 0)
  {
    errorMessage("Please start sntm as root.\n");
    return -1;
  }

  signalHandler();

  memset(&tMonitSrv, 0, sizeof(MONITSRVR));

  if(initMonitServer(pMntSrv))
  {
    errorMessage("initMonitServer fail ...\n");
    return -1;
  }
  
  pCnfgFile = calloc(1, strlen(CONFIG_PATH)+strlen("mconfig.xml")+2);

  if(!pCnfgFile)
  {
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
    errorMessage("%s\n", strerror(errno));
#else
    errorMessage("calloc failed.\n");
#endif

    return -1;
  }
  else
    sprintf(pCnfgFile, "%s/%s", CONFIG_PATH, "mconfig.xml");

  if(loadConfig(pMntSrv, pCnfgFile))
  {
    errorMessage("loadConfig fail ...\n");
    return -1;
  }

  if(!(pRootCa = getCnfgCharElement("cafile", pMntSrv->pSrvrCnfg)))
  {
    errorMessage("locate root CA fail ...\n");
    return -1;
  }

  if(!(pCaPrvKey = getCnfgCharElement("caprvkey", pMntSrv->pSrvrCnfg)))
  {
    errorMessage("locate private key file fail ...\n");
    return -1;
  }

  iPort = getCnfgIntElement("sensor_client", pMntSrv->pSrvrCnfg);
  if(iPort <=0 || iPort >0xFFFF)
    iPort = DFTSNSRPORT;

  sslInit();

  if(!(pMntSrv->pCert = (void *)loadCaCert(pRootCa, pCaPrvKey)))
  {
    errorMessage("initGenCert fail ...\n");
    return -1;
  }

  if((iSockFd = openSocket(iPort, 5)) == -1)
  {
    errorMessage("openSocket fail ...\n");
    return -1;
  }

  pMntSrv->tMainThdId = gmlThreadId();

  pMntSrv->tExtThrdId = gmlThreadCreate(doExtractThread, (void *)pMntSrv);

  pMntSrv->tDbiThrdId = gmlThreadCreate(doDbInsertThread, (void *)pMntSrv);

  /* Main loop to accept client connection */
  while(!pMntSrv->cException)
  {
    int iNewSock = -1;
    struct sockaddr tRmtAddr;
    socklen_t addrLen = sizeof(struct sockaddr);
    unsigned long ulAcceptedIP = 0;
    GDATAPOLL *pCrntClnt = NULL;

    debugMessage("Accepting connection ...\n");

    while((iNewSock = accept(iSockFd, &tRmtAddr, &addrLen)) == -1)
    {
      if(errno == EWOULDBLOCK || errno == EAGAIN)
      {
        /* No traffic */
        if(pMntSrv->cException)
          break;

        uSockWait(iSockFd, 1000000);
      }
      else
      {
        pMntSrv->cException = 1;
        errorMessage("%s\n", strerror(errno));
        break;
      }
    }

    ulAcceptedIP = ntohl(((struct sockaddr_in *)&tRmtAddr)->sin_addr.s_addr); 

    gmlMxLock(pMntSrv->pIfMtx);

    GETFIRST(pMntSrv->pClntInfo, pCrntClnt);
#define CLNTINFO pCrntClnt->pNext->pAny
    while(pCrntClnt->pNext && CLNTINFO)
    {
      CLIENTINF *pCrnt = (CLIENTINF *)CLNTINFO;
#undef CLNTINFO

      if(pCrnt->ulRmtIP == ulAcceptedIP) /* Address already in use. */
      {
        debugMessage("Sensor with this address has already connected ...\n");
        close(iNewSock);
        iNewSock = -1;
        break;
      }

      pCrntClnt=pCrntClnt->pNext;
    }

    gmlMxUnLock(pMntSrv->pIfMtx);

    while(!pMntSrv->cException && iNewSock != -1)
    {
      CLIENTINF *pClntInfo = NULL;

      debugMessage("Handle the new connection ...\n");

      if(!(pClntInfo = (CLIENTINF *)calloc(1, sizeof(CLIENTINF))))
      {
        pMntSrv->cException = 1;

#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
        errorMessage("%s\n", strerror(errno));
#else
        errorMessage("calloc failed.\n");
#endif
        break;
      }

      pClntInfo->iSockFd = iNewSock;
      pClntInfo->ulRmtIP = ulAcceptedIP;
      pClntInfo->tThrdId = gmlThreadCreate(doClientThread, (void *)pClntInfo);

      if(pClntInfo->tThrdId == -1)
      {
        free(pClntInfo);
        pMntSrv->cException = 1;
        errorMessage("gmlThreadCreate fail ...\n");
      }

      gmlThreadDetach(pClntInfo->tThrdId);

      break;
    } /* while */
  } /* while */

  close(iSockFd);
  freeCaCert(pMntSrv->pCert);
  dstryMonitServer(pMntSrv);
  free(pCnfgFile);

  return 0;
}
