/*
 * Copyright (C) 2001 - 2003 Min Hsu
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/x509.h>

#include "printmsg.h"

#define CHAINDEPTH  1  /* Certificate veirfication chain depth */

unsigned int uiGsidCtx = 0; /* Global context seesion ID */

void sslInit(void)
{
  SSL_library_init();
  SSL_load_error_strings();
}

void sslClean(void)
{
  ERR_remove_state(0);
  ERR_free_strings();
  EVP_cleanup();
}

static int verifyCert(int iOk, X509_STORE_CTX *pCtx)
{
  X509 *pErrCert = NULL;
  short sErr = 0, sDepth = 0;
  char *pMsg = NULL;

  if(!pCtx) /* Should not be NULL! */
    return 0;

  pErrCert = X509_STORE_CTX_get_current_cert(pCtx);
  sErr = X509_STORE_CTX_get_error(pCtx);
  sDepth=  X509_STORE_CTX_get_error_depth(pCtx);

#ifdef DEBUG
  pMsg = X509_NAME_oneline (X509_get_issuer_name(pErrCert),0,0);
  debugMessage("Issuer: %s\n", pMsg);
  free(pMsg);

  pMsg = X509_NAME_oneline(X509_get_subject_name(pErrCert),0, 0);
  debugMessage("Subject: %s\n", pMsg);
  free(pMsg);
#endif

  if(iOk)
  {
    debugMessage("Verify error: %d:%s\n", sErr,
                            X509_verify_cert_error_string(sErr));
    if(sDepth <= CHAINDEPTH)
      iOk=1;
    else
    {
      iOk=0;
      X509_STORE_CTX_set_error(pCtx, X509_V_ERR_CERT_CHAIN_TOO_LONG);
    }
  }

  switch (pCtx->error)
  {
    case X509_V_ERR_CERT_CHAIN_TOO_LONG:
         errorMessage("Certificate chain is too long!\n");
         break;

    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
         pMsg = X509_NAME_oneline(
                X509_get_issuer_name(pCtx->current_cert), 0 ,0);
         errorMessage("Unable to get CA: %s\n", pMsg);
         free(pMsg);
         iOk=0;
         break;

    case X509_V_ERR_CERT_NOT_YET_VALID:
    case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
         errorMessage("Certificate not valid yet!\n");
         iOk=0;
         break;

    case X509_V_ERR_CERT_HAS_EXPIRED:
    case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
         errorMessage("Certificate has expired!\n");
         iOk=0;
         break;
  }

  return(iOk);
}

SSL_CTX *initServerCtx(X509 *pCert, EVP_PKEY *pPkey)
{
  SSL_CTX *pCtx = NULL;
  SSL_METHOD *pMthd = NULL;

  pMthd = SSLv3_server_method();

  pCtx = SSL_CTX_new(pMthd);
  if(!pCtx)
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  SSL_CTX_set_options(pCtx, SSL_OP_NO_SSLv2 | SSL_OP_NO_TLSv1);

  SSL_CTX_set_verify(pCtx, SSL_VERIFY_NONE, NULL);
  SSL_CTX_set_session_id_context(pCtx,(void*)&uiGsidCtx, sizeof(uiGsidCtx));

  if(SSL_CTX_use_certificate(pCtx, pCert) <= 0)
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  if(SSL_CTX_use_PrivateKey(pCtx, pPkey) <= 0)
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  if(!SSL_CTX_check_private_key(pCtx))
  {
    errorMessage("%s\n", "Private key unmatch the certificate public key");
    return NULL;
  }

  return pCtx;
}

SSL_CTX *initClientCtx(void *pCert)
{
  SSL_CTX *pCtx = NULL;
  SSL_METHOD *pMthd = NULL;

  if(!pCert)
    return NULL;

  pMthd = SSLv3_client_method();

  pCtx = SSL_CTX_new(pMthd);
  if(!pCtx)
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  SSL_CTX_set_options(pCtx, SSL_OP_NO_SSLv2 | SSL_OP_NO_TLSv1);

  SSL_CTX_set_verify_depth(pCtx, CHAINDEPTH);
  SSL_CTX_set_verify(pCtx, SSL_VERIFY_PEER , verifyCert);

  if(!X509_STORE_add_cert(pCtx->cert_store, pCert))
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 

  return pCtx;
}

void cleanCtx(SSL_CTX *pCtx)
{
  if(pCtx)
    SSL_CTX_free(pCtx);
}

SSL *acceptClient(SSL_CTX *pCtx, const int sckt)
{
  SSL *pConnect = NULL;

  if(!pCtx || sckt <= 0)
    return NULL;

  if(!(pConnect = SSL_new(pCtx)))
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  if(SSL_set_fd(pConnect, sckt) > 0)
  {
    if(SSL_accept(pConnect) > 0) /* Accept from client */
      return pConnect;
    else
    {
      if(ERR_get_error())
        errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
      else
        errorMessage("%s\n", strerror(errno));
    }
  }
  else
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
        
  SSL_free (pConnect);
  return NULL;
}

SSL *connetServer(SSL_CTX *pCtx, const int sckt)
{
  SSL *pConnect = NULL;

  if(!pCtx || sckt <= 0)
    return NULL;

  if(!(pConnect = SSL_new(pCtx)))
  {
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
    return NULL;
  }

  if(SSL_set_fd(pConnect, sckt) > 0)
  {
    if(SSL_connect(pConnect) > 0) /* Connect to server */
      return pConnect;
    else
    {
      if(ERR_get_error())
        errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
      else
        errorMessage("%s\n", strerror(errno));
    }
  }
  else
    errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
        
  SSL_free (pConnect);
  return NULL;
}

void disconnectSSL(SSL *pConnect, const short sShut)
{
  if(pConnect)
  {
    SSL_clear(pConnect);

    if(sShut)
      SSL_shutdown(pConnect);
    else
      SSL_set_shutdown(pConnect, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);

    SSL_free(pConnect);
  }
}

int readSSL(SSL *pConnect, unsigned char *pData, const int iLen)
{
  int iDataLen = iLen;
  int iDataRcv = 0;
  int iRcvLen = 0;

  if(!pConnect || !pData)
    return -1;

  while(iDataLen > 0)
  {
    iDataRcv = SSL_read(pConnect, pData+iRcvLen, iLen);        

    switch(SSL_get_error(pConnect, iDataRcv))
    {
      case SSL_ERROR_NONE:
             if(SSL_pending(pConnect))
             {
               iRcvLen += iDataRcv;
               iDataLen -= iDataRcv;
               break;
             }
             else
               return iDataRcv;             

      case SSL_ERROR_WANT_READ:
      case SSL_ERROR_WANT_WRITE:
      case SSL_ERROR_WANT_X509_LOOKUP:
             usleep(1000);
             break;

      default:
             errorMessage("%d\n", SSL_get_error(pConnect, iDataRcv)); 
             return -1;
    }
  }

  return iRcvLen;
}

int writeSSL(SSL *pConnect, const unsigned char *pData, const int iLen)
{
  int iDataLen = iLen;
  int iDataSnd = 0;
  int iSndLen = 0;

  if(!pConnect || !pData)
    return -1;

  while(iDataLen > 0)
  {
    iDataSnd = SSL_write(pConnect, pData+iSndLen, iLen);
    switch(SSL_get_error(pConnect, iDataSnd))
    {
      case SSL_ERROR_NONE:
             iSndLen += iDataSnd;
             iDataLen -= iDataSnd;
             break;
      case SSL_ERROR_WANT_WRITE:
      case SSL_ERROR_WANT_READ:
      case SSL_ERROR_WANT_X509_LOOKUP:
             usleep(1000);
             break;
      default:
             errorMessage("%s\n", ERR_reason_error_string(ERR_get_error())); 
             return -1;
    }
  }
   
  return iSndLen;
}
